@@ -181,7 +181,7 @@ def attach_to_group(group):
 
     @repo.command()
     @click.argument("url")
-    @click.option("--clients-auth-path", default=None, help="Directory where authentication repository is located.")
+    @click.argument("clients-auth-path", default=None, required=False)
     @click.option("--clients-library-dir", default=None, help="Directory where target repositories and, "
                   "optionally, authentication repository are located. If omitted it is "
                   "calculated based on authentication repository's path. "
@@ -200,8 +200,10 @@ def attach_to_group(group):
                   "on if build was successful and error message if it was raised")
     @click.option("--exclude-target", multiple=True, help="globs defining which target repositories should be "
                   "ignored during update.")
+    @click.option("--strict", is_flag=True, default=False, help="Enable/disable strict mode - return an error"
+                  "if warnings are raised ")
     def update(url, clients_auth_path, clients_library_dir, default_branch, from_fs, expected_repo_type,
-               scripts_root_dir, profile, format_output, exclude_target):
+               scripts_root_dir, profile, format_output, exclude_target, strict):
         """
         Update and validate local authentication repository and target repositories. Remote
         authentication's repository url needs to be specified when calling this command. If the
@@ -209,13 +211,17 @@ def attach_to_group(group):
         locations of the target repositories are calculated based on the authentication repository's
         path. If that is not the case, it is necessary to redefine this default value using the
         --clients-library-dir option. This means that if authentication repository's path is
-        E:\\root\\namespace\\auth-repo, it will be assumed that E:\\root is the root direcotry
+        E:\\root\\namespace\\auth-repo, it will be assumed that E:\\root is the root directory
         if clients-library-dir is not specified.
         Names of target repositories (as defined in repositories.json) are appended to the root repository's
         path thus defining the location of each target repository. If names of target repositories
         are namespace/repo1, namespace/repo2 etc and the root directory is E:\\root, path of the target
         repositories will be calculated as E:\\root\\namespace\\repo1, E:\\root\\namespace\\root2 etc.
 
+        Path to authentication repository directory is set by clients-auth-path argument. If this
+        argument is not provided, it is expected that --clients-library-dir option is used. The updater
+        will raise an error if one of both isn't set.
+
         If remote repository's url is a file system path, it is necessary to call this command with
         --from-fs flag so that url validation is skipped. When updating a test repository (one that has
         the "test" target file), use --authenticate-test-repo flag. An error will be raised
@@ -228,6 +234,9 @@ def attach_to_group(group):
         One or more target repositories can be excluded from the update process using --exclude-target.
         In that case, the library will only be partly validated, so last_validate_commit will not be updated
         and no scripts will be called.
+
+        Update can be in strict or no-strict mode. Strict mode is set by specifying --strict, which will raise errors
+        during update if any/all warnings are found. By default, --strict is disabled.
         """
         if clients_auth_path is None and clients_library_dir is None:
             raise click.UsageError('Must specify either authentication repository path or library directory!')
@@ -257,7 +266,8 @@ def attach_to_group(group):
                 from_fs,
                 UpdateType(expected_repo_type),
                 scripts_root_dir=scripts_root_dir,
-                excluded_target_globs=exclude_target
+                excluded_target_globs=exclude_target,
+                strict=strict
             )
             if format_output:
                 print(json.dumps({
@@ -281,10 +291,17 @@ def attach_to_group(group):
                   "Authentication repo is presumed to be at library-dir/namespace/auth-repo-name")
     @click.option("--default-branch", default="main", help="Name of the default branch, like mian or master")
     @click.option("--from-commit", default=None, help="First commit which should be validated.")
-    def validate(clients_auth_path, clients_library_dir, default_branch, from_commit):
+    @click.option("--exclude-target", multiple=True, help="globs defining which target repositories should be "
+                  "ignored during update.")
+    @click.option("--strict", is_flag=True, default=False, help="Enable/disable strict mode - return an error"
+                  "if warnings are raised")
+    def validate(clients_auth_path, clients_library_dir, default_branch, from_commit, exclude_target, strict):
         """
         Validates an authentication repository which is already on the file system
         and its target repositories (which are also expected to be on the file system).
         Does not clone repositories, fetch changes or merge commits.
+
+        Validation can be in strict or no-strict mode. Strict mode is set by specifying --strict, which will raise errors
+        during validate if any/all warnings are found. By default, --strict is disabled.
         """
-        validate_repository(clients_auth_path, clients_library_dir, default_branch, from_commit)
+        validate_repository(clients_auth_path, clients_library_dir, default_branch, from_commit, exclude_target, strict)
