@@ -2,14 +2,14 @@ import json
 import shutil
 import enum
 import tempfile
-import tuf
-import tuf.client.updater as tuf_updater
+
+
+from tuf.ngclient.updater import Updater
 from tuf.repository_tool import TARGETS_DIRECTORY_NAME
 
 from collections import defaultdict
 from pathlib import Path
 from taf.log import taf_logger, disable_tuf_console_logging
-from taf.git import GitRepository
 import taf.repositoriesdb as repositoriesdb
 from taf.auth_repo import AuthenticationRepository
 from taf.utils import timed_run
@@ -20,6 +20,7 @@ from taf.exceptions import (
     GitError,
     MissingHostsError,
     InvalidHostsError,
+    ValidationFailedError,
 )
 from taf.updater.handlers import GitUpdater
 from taf.utils import on_rm_error
@@ -37,6 +38,7 @@ from taf.updater.lifecycle_handlers import (
 )
 from cattr import unstructure
 
+EXPIRED_METADATA_ERROR = "ExpiredMetadataError"
 PROTECTED_DIRECTORY_NAME = "protected"
 INFO_JSON_PATH = f"{TARGETS_DIRECTORY_NAME}/{PROTECTED_DIRECTORY_NAME}/info.json"
 
@@ -82,7 +84,7 @@ def _clone_validation_repo(url, repository_name, default_branch):
     """
     temp_dir = tempfile.mkdtemp()
     path = Path(temp_dir, "auth_repo").absolute()
-    validation_auth_repo = GitRepository(path=path, urls=[url])
+    validation_auth_repo = AuthenticationRepository(path=path, urls=[url])
     validation_auth_repo.clone(bare=True)
     validation_auth_repo.fetch(fetch_all=True)
 
@@ -96,7 +98,7 @@ def _clone_validation_repo(url, repository_name, default_branch):
             repository_name = f'{info["namespace"]}/{info["name"]}'
         except Exception:
             raise UpdateFailedError(
-                "Error during info.json parse. Check if info.json exists or provide full path to auth repo"
+                "Error during info.json parse. When specifying --clients-library-dir check if info.json metadata exists in targets/protected or provide full path to auth repo"
             )
 
     validation_auth_repo.cleanup()
@@ -202,6 +204,7 @@ def update_repository(
     scripts_root_dir=None,
     checkout=True,
     excluded_target_globs=None,
+    strict=False,
 ):
     """
     <Arguments>
@@ -225,11 +228,13 @@ def update_repository(
         See repositoriesdb load_repositories for more details.
     checkout:
         Whether to checkout last validated commits after update is done
+    strict:
+        Whether or not update fails if a warning is raised.
     """
+    settings.strict = strict
     # if the repository's name is not provided, divide it in parent directory
     # and repository name, since TUF's updater expects a name
     # but set the validate_repo_name setting to False
-
     if clients_auth_path is not None:
         clients_auth_path = Path(clients_auth_path).resolve()
 
@@ -271,7 +276,9 @@ def update_repository(
             excluded_target_globs=excluded_target_globs,
         )
     except Exception as e:
-        root_error = e
+        root_error = UpdateFailedError(
+            f"Update of {auth_repo_name} failed due to error {e}"
+        )
 
     update_data = {}
     if not excluded_target_globs:
@@ -422,7 +429,7 @@ def _update_named_repository(
     that TUF skips mirrors which do not have valid and/or current metadata files. Also, we
     do not simply want to find the latest metadata, we want to validate everything in-between.
     That is why the idea is to call refresh multiple times, until the last commit is reached.
-    The 'GitMetadataUpdater' updater is designed in such a way that for each new call it
+    The 'GitUpdater' updater is designed in such a way that for each new call it
     loads data from a most recent commit.
     """
     if visited is None:
@@ -600,19 +607,6 @@ def _update_current_repository(
     settings.update_from_filesystem = update_from_filesystem
     settings.conf_directory_root = conf_directory_root
     settings.default_branch = default_branch
-    if only_validate:
-        settings.overwrite_last_validated_commit = True
-        settings.last_validated_commit = validate_from_commit
-    # instantiate TUF's updater
-    repository_mirrors = {
-        "mirror1": {
-            "url_prefix": url,
-            "metadata_path": "metadata",
-            "targets_path": "targets",
-            "confined_target_dirs": [""],
-        }
-    }
-    tuf.settings.repositories_directory = clients_auth_library_dir
 
     def _commits_ret(commits, existing_repo, update_successful):
         if commits is None:
@@ -635,11 +629,6 @@ def _update_current_repository(
 
     try:
         commits = None
-        users_repo_existed = False
-
-        # first clone the validation repository in temp. this is needed because tuf expects auth_repo_name to be valid (not None)
-        # and in the right format (seperated by '/'). this approach covers a case where we don't know authentication repo path upfront.
-        auth_repo_name = _clone_validation_repo(url, auth_repo_name, default_branch)
         # check whether the directory that runs clone exists or contains additional files.
         # we need to check the state of folder before running tuf. Resolves issue #22
         # if auth_repo_name isn't specified then the current directory doesn't contain additional files.
@@ -648,10 +637,11 @@ def _update_current_repository(
             if auth_repo_name is not None
             else True
         )
-
-        repository_updater = tuf_updater.Updater(
-            auth_repo_name, repository_mirrors, GitUpdater
-        )
+        # first clone the validation repository in temp. this is needed because tuf expects auth_repo_name to be valid (not None)
+        # and in the right format (seperated by '/'). this approach covers a case where we don't know authentication repo path upfront.
+        auth_repo_name = _clone_validation_repo(url, auth_repo_name, default_branch)
+        git_updater = GitUpdater(url, clients_auth_library_dir, auth_repo_name)
+        _run_tuf_updater(git_updater)
     except Exception as e:
         # Instantiation of the handler failed - this can happen if the url is not correct
         # of if the saved last validated commit does not match the current head commit
@@ -682,7 +672,7 @@ def _update_current_repository(
         )
     try:
 
-        users_auth_repo = repository_updater.update_handler.users_auth_repo
+        users_auth_repo = git_updater.users_auth_repo
         existing_repo = users_auth_repo.is_git_repository_root
 
         (
@@ -690,7 +680,7 @@ def _update_current_repository(
             error_msg,
             last_validated_commit,
         ) = _validate_authentication_repository(
-            repository_updater,
+            git_updater,
             users_auth_repo,
             out_of_band_authentication,
             auth_repo_name,
@@ -738,7 +728,6 @@ def _update_current_repository(
             checkout,
         )
     except Exception as e:
-
         if not existing_repo:
             shutil.rmtree(users_auth_repo.path, onerror=on_rm_error)
             shutil.rmtree(users_auth_repo.conf_dir)
@@ -764,59 +753,6 @@ def _update_current_repository(
     )
 
 
-def _update_authentication_repository(repository_updater):
-
-    users_auth_repo = repository_updater.update_handler.users_auth_repo
-    taf_logger.info("Validating authentication repository {}", users_auth_repo.name)
-    try:
-        # for each commit, load that revision's metadata and target files
-        # the implementation is based on examples listed in TUF's documentation
-        # (/tuf/client/README.md)
-        while not repository_updater.update_handler.update_done():
-            current_commit = repository_updater.update_handler.current_commit
-            repository_updater.refresh()
-            # using refresh, we have updated all main roles
-            # we still need to update the delegated roles (if there are any)
-            # that is handled by get_one_valid_targetinfo
-            current_targets = repository_updater.update_handler.get_current_targets()
-            taf_logger.debug("Validated metadata files at revision {}", current_commit)
-            for target_path in current_targets:
-                target = repository_updater.get_one_valid_targetinfo(target_path)
-                target_filepath = target["filepath"]
-                trusted_length = target["fileinfo"]["length"]
-                trusted_hashes = target["fileinfo"]["hashes"]
-                try:
-                    # just call get target to validate it instead of callind download_target
-                    repository_updater._get_target_file(
-                        target_filepath, trusted_length, trusted_hashes
-                    )
-                except tuf.exceptions.NoWorkingMirrorError as e:
-                    taf_logger.error("Could not validate file {}", target_filepath)
-                    raise e
-                taf_logger.debug(
-                    "Successfully validated target file {} at {}",
-                    target_filepath,
-                    current_commit,
-                )
-    except Exception as e:
-        # for now, useful for debugging
-        taf_logger.error(
-            "Validation of authentication repository {} failed at revision {} due to error {}",
-            users_auth_repo.name,
-            current_commit,
-            e,
-        )
-
-        raise UpdateFailedError(
-            f"Validation of authentication repository {users_auth_repo.name}"
-            f" failed at revision {current_commit} due to error: {e}"
-        )
-
-    taf_logger.info(
-        "Successfully validated authentication repository {}", users_auth_repo.name
-    )
-
-
 def _update_target_repositories(
     repositories,
     repositories_branches_and_commits,
@@ -978,6 +914,72 @@ def _set_target_old_head_and_validate(
     return old_head
 
 
+def _run_tuf_updater(git_updater):
+    def _init_updater():
+        try:
+            return Updater(
+                git_updater.metadata_dir,
+                "metadata/",
+                git_updater.targets_dir,
+                "targets/",
+                fetcher=git_updater,
+            )
+        except Exception as e:
+            taf_logger.error(f"Failed to instantiate TUF Updater due to error {e}")
+            raise e
+
+    def _update_tuf_current_revision():
+        current_commit = git_updater.current_commit
+        try:
+            updater.refresh()
+            taf_logger.debug("Validated metadata files at revision {}", current_commit)
+            # using refresh, we have updated all main roles
+            # we still need to update the delegated roles (if there are any)
+            # and validate any target files
+            current_targets = git_updater.get_current_targets()
+            for target_path in current_targets:
+                target_filepath = target_path.replace("\\", "/")
+
+                targetinfo = updater.get_targetinfo(target_filepath)
+                target_data = git_updater.get_current_target_data(
+                    target_filepath, raw=True
+                )
+                targetinfo.verify_length_and_hashes(target_data)
+
+                taf_logger.debug(
+                    "Successfully validated target file {} at {}",
+                    target_filepath,
+                    current_commit,
+                )
+        except Exception as e:
+            metadata_expired = EXPIRED_METADATA_ERROR in type(
+                e
+            ).__name__ or EXPIRED_METADATA_ERROR in str(e)
+            if not metadata_expired or settings.strict:
+                taf_logger.error(
+                    "Validation of authentication repository {} failed at revision {} due to error {}",
+                    git_updater.users_auth_repo.name,
+                    current_commit,
+                    e,
+                )
+                raise UpdateFailedError(
+                    f"Validation of authentication repository {git_updater.users_auth_repo.name}"
+                    f" failed at revision {current_commit} due to error: {e}"
+                )
+            taf_logger.warning(
+                f"WARNING: Could not validate authentication repository {git_updater.users_auth_repo.name} at revision {current_commit} due to error {e}"
+            )
+
+    while not git_updater.update_done():
+        updater = _init_updater()
+        _update_tuf_current_revision()
+
+    taf_logger.info(
+        "Successfully validated authentication repository {}",
+        git_updater.users_auth_repo.name,
+    )
+
+
 def _get_commits(
     repository,
     existing_repository,
@@ -993,7 +995,9 @@ def _get_commits(
 
     if old_head is not None:
         if not only_validate:
-            fetched_commits = repository.all_fetched_commits(branch=branch)
+            fetched_commits = repository.all_commits_on_branch(
+                branch=f"origin/{branch}"
+            )
 
             # if the local branch does not exist (the branch was not checked out locally)
             # fetched commits will include already validated commits
@@ -1026,7 +1030,9 @@ def _get_commits(
             new_commits_on_repo_branch = []
         if not only_validate:
             try:
-                fetched_commits = repository.all_fetched_commits(branch=branch)
+                fetched_commits = repository.all_commits_on_branch(
+                    branch=f"origin/{branch}"
+                )
                 # if the local branch does not exist (the branch was not checked out locally)
                 # fetched commits will include already validated commits
                 # check which commits are newer that the previous head commit
@@ -1069,16 +1075,21 @@ def _merge_commit(repository, branch, commit_to_merge, checkout=True):
     taf_logger.info("Merging commit {} into {}", commit_to_merge, repository.name)
     try:
         repository.checkout_branch(branch, raise_anyway=True)
-    except GitError:
-        # in order to merge a commit into a branch we need to check it out
-        # but that cannot be done if it is checked out in a different worktree
-        # it should be fine to update that other worktree if there are no uncommitted changes
-        # or no commits that have not been pushed yet
-        worktree = GitRepository(path=repository.find_worktree_path_by_branch(branch))
-        if worktree is None:
-            return False
-        repository = worktree
-        checkout = False
+    except GitError as e:
+        # two scenarios:
+        # current git repository is in an inconsistent state:
+        # - .git/index.lock exists (git partial update got applied)
+        # should get addressed in https://github.com/openlawlibrary/taf/issues/210
+        # current git repository has uncommitted changes:
+        # we do not want taf to lose any repo data, so we do not reset the repository.
+        # for now, raise an update error and let the user manually reset the repository
+        taf_logger.error(
+            "Could not checkout branch {} during commit merge. Error {}", branch, e
+        )
+        raise UpdateFailedError(
+            f"Repository {repository.name} should contain only committed changes. \n"
+            + f"Please update the repository at {repository.path} manually and try again."
+        )
 
     repository.merge_commit(commit_to_merge)
     if checkout:
@@ -1236,7 +1247,10 @@ def validate_repository(
     clients_library_dir=None,
     default_branch="main",
     validate_from_commit=None,
+    excluded_target_globs=None,
+    strict=False,
 ):
+    settings.strict = strict
 
     clients_auth_path = Path(clients_auth_path).resolve()
 
@@ -1252,17 +1266,27 @@ def validate_repository(
         if (clients_auth_path / "targets" / "test-auth-repo").exists()
         else UpdateType.OFFICIAL
     )
-    _update_named_repository(
-        str(clients_auth_path),
-        clients_auth_library_dir,
-        clients_library_dir,
-        auth_repo_name,
-        default_branch,
-        True,
-        expected_repo_type=expected_repo_type,
-        only_validate=True,
-        validate_from_commit=validate_from_commit,
-    )
+    settings.overwrite_last_validated_commit = True
+    settings.last_validated_commit = validate_from_commit
+    try:
+        _update_named_repository(
+            str(clients_auth_path),
+            clients_auth_library_dir,
+            clients_library_dir,
+            auth_repo_name,
+            default_branch,
+            True,
+            expected_repo_type=expected_repo_type,
+            only_validate=True,
+            validate_from_commit=validate_from_commit,
+            excluded_target_globs=excluded_target_globs,
+        )
+    except Exception as e:
+        raise ValidationFailedError(
+            f"Validation or repository {auth_repo_name} failed due to error {e}"
+        )
+    settings.overwrite_last_validated_commit = False
+    settings.last_validated_commit = None
 
 
 def _validate_authentication_repository(
@@ -1273,17 +1297,10 @@ def _validate_authentication_repository(
     expected_repo_type,
 ):
     error_msg = None
-    # validate the authentication repository and fetch new commits
-    # if the validation is completed successfully, new commits are fetched (not merged yet)
-    try:
-        _update_authentication_repository(repository_updater)
-    except Exception as e:
-        error_msg = e
-
     # this is the repository cloned inside the temp directory
     # we validate it before updating the actual authentication repository
-    validation_auth_repo = repository_updater.update_handler.validation_auth_repo
-    commits = repository_updater.update_handler.commits
+    validation_auth_repo = repository_updater.validation_auth_repo
+    commits = repository_updater.commits
 
     if (
         out_of_band_authentication is not None
@@ -1302,22 +1319,22 @@ def _validate_authentication_repository(
 
     if expected_repo_type != UpdateType.EITHER:
         # check if the repository being updated is a test repository
-        targets = validation_auth_repo.get_json(commits[-1], "metadata/targets.json")
-        test_repo = "test-auth-repo" in targets["signed"]["targets"]
-        if test_repo and expected_repo_type != UpdateType.TEST:
+        if validation_auth_repo.is_test_repo and expected_repo_type != UpdateType.TEST:
             error_msg = UpdateFailedError(
                 f"Repository {users_auth_repo.name} is a test repository. "
                 'Call update with "--expected-repo-type" test to update a test '
                 "repository"
             )
-        elif not test_repo and expected_repo_type == UpdateType.TEST:
+        elif (
+            not validation_auth_repo.is_test_repo
+            and expected_repo_type == UpdateType.TEST
+        ):
             error_msg = UpdateFailedError(
                 f"Repository {users_auth_repo.name} is not a test repository,"
                 ' but update was called with the "--expected-repo-type" test'
             )
-
     # always cleanup repository updater
-    repository_updater.update_handler.cleanup()
+    repository_updater.cleanup()
 
     return (
         commits,
